package org.purl.jh.esslli.data;

import com.google.common.base.Preconditions;
import com.google.common.base.Splitter;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

/**
 * A simple rule specifying the change of a string's tail.
 * The rule contains a sequence of A -&gt B sub-rules. The first one to 
 * matching the transformaed string (i.e. the string ends in A) is used and the 
 * transformation ends. If no sub-rule matches the string, not change is performed.
 * 
 * Implementing this as an fs transducer would be much more efficient.
 */
@lombok.Data
public final class TailChangeRule implements StringRule {
    private final ImmutableList<String> oldTails;
    private final ImmutableList<String> newTails;
    private final ImmutableList<Boolean> opt;
    
    /**
     * Creates a new rule based on a string specification.
     *
     * @param aRuleSpec rule specification in the following form:
     * <pre>
     * rule = subrule(:subrule)*
     * subrule = string[?>]string
     * </pre>
     * Where &gt; marks obligatory change and ? marks optional change.
     * 
     * For example:
     * <pre>
     * a>b:b?c
     * </pre>
     */
    public static TailChangeRule fromSpecString(String str) {
        Preconditions.checkArgument(str.matches("[^?>]*[?>][^?>]*(:[^?>]*[?>][^?>])*"), "Incorrect change rule format %s", str);

        final List<String> oldTails = new ArrayList<>();
        final List<String> newTails = new ArrayList<>();
        final List<Boolean> opt = new ArrayList<>();

        final Splitter obligSplitter = Splitter.on('>'); // splits "a>b" subrule to ["a", "b"]
        final Splitter optSplitter   = Splitter.on('?'); // splits "a?b" subrule to ["a", "b"]

        // split into subrules and parse each of them 
        for (String subrule : Splitter.on(':').split(str)) {
            Splitter splitter;
            if (subrule.contains(">")) {
                opt.add(false);
                splitter = obligSplitter;
            }
            else {
                opt.add(true);
                splitter = optSplitter;
            }

            Iterator<String> it = splitter.split(subrule).iterator();
            oldTails.add(it.next());
            newTails.add(it.next());
        }
        
        return new TailChangeRule(ImmutableList.copyOf(oldTails), ImmutableList.copyOf(newTails), ImmutableList.copyOf(opt));
    }
    
    @Override
    public Set<String> apply(String input) {
        for (int i = 0; i < oldTails.size(); i++) {
            if ( input.endsWith(oldTails.get(i)) ) {
                String newString = replaceTail(input, oldTails.get(i), newTails.get(i));

                // returned the changed string. If the change was optional return the original as well 
                return opt.get(i) ? ImmutableSet.of(input, newString) : ImmutableSet.of(newString);
            }
        }

        // return the original string as now change could be applied
        return ImmutableSet.of(input);      
    }

    /**
     * Removes oldTail from the end of input and replaces it with newTail. No checks are performed.
     */
    protected static String replaceTail(String input, String oldTail, String newTail) {
        return input.substring(0, input.length()-oldTail.length()) + newTail;
    }
}
